const {NOT,PICK,pipe} = require('./func-utils');

const getPageProcessor = require('./getPageProcessor');


/**
 * Convert a list of lines into a list of sections
 *
 *  where a section is an array [type, string-of-lines]
 *     {type} is one of '!head', '!md', '!html', '!tryit', '!end'
 *     {string-of-lines} is a string with linex seperated by the new-line character '\n'
 *     
 * @param  {[string]} lines list of lines
 * @return {[[string,string]}       see description above
 */
module.exports.sections = function (_lines) {
    // console.log(_lines);
    let lines = _lines.flatMap(getMarkers());
    let lastLine = '';
    const processPages = getPageProcessor();
    var [list, type, content] = lines.reduce( ([list, type, content], line) =>{
        if(line.match(/^(![a-z]+|!--)/)) {
           if(content) list.push([type, content]);
           lastLine="";
           return [list, line, ''];
        }
        if(line === '' && lastLine === '') {
            line = '';
        } else {
            lastLine = line.trim();
            line = (content?'\n':'')+line;
        }
        return [list, type, content+line];
    }, [[], '!html', '']);
    if(type !== '!end' || content) list.push([type, content]);
    return makePages(list,processPages);
} ;

function truePath(path) {
    path = path.replace(/\/+/g,'/').replace(/\\+/g, '/').replace(/_/g, '&#95;');
    let pa = path.split('/');
    let res = [pa[0]];
    for(let i=1; i<pa.length; i++) {
        let e = pa[i];
        if(e === '.') continue;
        if(e === '..') {
            res.pop();
        } else res.push(e);
    }
    return res.join('/');
}

function removePrefixStr() {
    let prev = '';
    return ((current) => {
        let pl = prev.split('/');
        let cl = current.split('/');
        prev = current;
        if( pl.length !== cl.length) return current;
        for(let i=0; i<pl.length; i++) {
            if(pl[i] !== cl[i]) return cl.slice(i).join('/');
        }
        return current;
    });
}


class Stack {
    constructor() {
        this.stack = [];
        this.state = undefined;
        this._path = '';
    }

    push(v) {
        let name = this.name(v);
        if(v.trim().startsWith('<!-- @@]')) {
            if(this.top() !== name) throw new Error('invalid INCLUDE NAME: ' + name + " expected name: "+this.top());
            // console.log('pop',v);
            this.stack.pop();
            return [];
        }
        else if(v.trim().startsWith('<!-- @@[')) {
            if(!name) throw new Error('invalid INCLUDE NAME: ' + v);
            // console.log('push',v);
            this.stack.push(name);
            return [];
        }
        if(v.startsWith('!')) this.state = v.trim();
        else if(this.state === '!md') {
            if(v.trim()[0] === '#') {
                let _path = this.path();
                let color = this._path === _path ? 'grey' : 'circular blue';
                this._path = _path;
                // console.log(`<i class="icon sitemap green icon-right" title="Include path: ${this.path()}"></i>`, this.stack);
                return [v, `<i class="icon tiny sitemap ${color} icon-right" title="Include path: ${_path}"></i>`]
            }
        }
        return v;
    }
    name(str) {
        let nameV = str.match(/<!-- @@[\]\[] (.*) -->/);
        return nameV && truePath(nameV[1]);
    }

    path() {
        //return this.stack.join(' ➯ ');
        return this.stack.map(removePrefixStr()).reverse().join(' < ');
    }

    top() {
        if(this.stack.lenght === 0) return undefined;
        return this.stack[this.stack.length-1];
    }
}


function getMarkers() {
    let stack = new Stack()
    return (e => stack.push(e))
}



/**
 * Precess all the section and if
 *
 * section type is not '!md' leave it as it
 * if it is a '!md' section use the '#' (<h1>) heading to place a page marker, where 
 * each time be see <h1> marker that represents the end of the last page 
 * and the begining of the next
 * 
 * @param  {[section]} list     list of sections
 * @param  {function} pageProc page marker function
 * @return {[section]}          list of sections with page markers applied
 */
function makePages(list, pageProc) {
    let res =  list.map((elem) => {
        let [type, content] = elem;
        if( isMD(type) ) {  // if in a markdown 
            return [type, pageProc(content)]; // add page markers and remember the last !md section
        }
        else return elem;  // leave the non-markdown sections alone
    });
    const SECTION_TYPE = PICK(0); // section = [type, body_string]
    res = [
            ...res.filter( pipe(SECTION_TYPE, NOT(isEnd)) ), // get rid of any !end markers in the middle
           ['!html', pageProc()],  // terminate the last page (if there is any)
           ['!end', ''] //  finally add !end marker
    ];
    return res;
}


function isMD(type) {
    type = type.trim();
    return type === '!md' || type.startsWith('!md')
}

function isEnd(type) {
    type = type.trim();
    return type.startsWith('!end');
}